import contextlib
from dataclasses import dataclass
from functools import partial
from typing import Any, Dict, Type, TypeVar

from pydantic import VERSION, BaseModel
from typing_extensions import TypeGuard

from zenconfig.base import BaseConfig, Schema
from zenconfig.encoder import Encoder, encode

C = TypeVar("C", bound=BaseModel)

PYDANTIC_V1 = VERSION.startswith("1.")


@dataclass
class PydanticSchema(Schema[C]):
    exclude_unset: bool = False
    exclude_defaults: bool = True

    def handles(self, cls: type) -> TypeGuard[Type[BaseModel]]:
        return issubclass(cls, BaseModel)

    def from_dict(self, cls: Type[C], cfg: Dict[str, Any]) -> C:
        if PYDANTIC_V1:
            return cls.parse_obj(cfg)
        return cls.model_validate(cfg)

    def to_dict(self, config: C, encoder: Encoder) -> Dict[str, Any]:
        # Use pydantic encoders.
        if PYDANTIC_V1:
            return _encoder(config)(
                config.dict(
                    exclude_unset=self.exclude_unset,
                    exclude_defaults=self.exclude_defaults,
                )
            )
        else:
            return config.model_dump(
                exclude_unset=self.exclude_unset,
                exclude_defaults=self.exclude_defaults,
                mode="json",
            )


BaseConfig.register_schema(PydanticSchema())


def _encoder(config: BaseModel) -> Encoder:
    def _enc(obj: Any) -> Any:
        with contextlib.suppress(TypeError):
            return config.__json_encoder__(obj)  # type: ignore
        return obj

    return partial(encode, _enc)



import logging
from abc import ABC
from functools import partial
from typing import ClassVar, Optional

from zenconfig.base import ZenConfigError
from zenconfig.encoder import Encoder, Encoders, combine_encoders, encode
from zenconfig.read import ReadOnlyConfig

logger = logging.getLogger(__name__)


class Config(ReadOnlyConfig, ABC):
    """Abstract base class for handling read and write operations."""

    # File mode used if we need to create the config file.
    FILE_MODE: ClassVar[int] = 0o600
    # Add custom encoders.
    ENCODERS: ClassVar[Encoders] = {}
    # Cached encoder.
    __ENCODER: ClassVar[Optional[Encoder]] = None

    def save(self) -> None:
        """Save the current config to the file."""
        paths = self._paths()
        if len(paths) != 1:
            raise ZenConfigError(
                "cannot save when handling multiple configuration files"
            )
        path = paths[0]
        if not path.exists():
            logger.debug("creating file at path %s", path)
            path.parent.mkdir(parents=True, exist_ok=True)
            path.touch(mode=self.FILE_MODE)
        fmt = self._format()
        logger.debug(
            "using %s to save %s to %s",
            fmt.__class__.__name__,
            self.__class__.__name__,
            path,
        )
        schema = self._schema()
        fmt.dump(path, schema.to_dict(self, self._encoder()))

    def clear(self) -> None:
        """Delete the config file(s)."""
        for path in self._paths():
            logger.debug("deleting file at path %s", path)
            path.unlink(missing_ok=True)

    @classmethod
    def _encoder(cls) -> Encoder:
        """Get the encoder, taking into account custom encoders."""
        if cls.__ENCODER:
            return cls.__ENCODER
        encoder = combine_encoders(cls.ENCODERS)
        cls.__ENCODER = partial(encode, encoder)
        return cls.__ENCODER



import contextlib

from zenconfig.formats.json import JSONFormat

with contextlib.suppress(ImportError):
    from zenconfig.formats.yaml import YAMLFormat
with contextlib.suppress(ImportError):
    from zenconfig.formats.toml import TOMLFormat



import contextlib

from zenconfig.schemas.dataclass import DataclassSchema
from zenconfig.schemas.dict import DictSchema

with contextlib.suppress(ImportError):
    from zenconfig.schemas.pydantic import PydanticSchema
with contextlib.suppress(ImportError):
    from zenconfig.schemas.attrs import AttrsSchema



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

from zenconfig.encoder import Encoder


class ZenConfigError(Exception):
    """Default error when Handling config files."""


class Format(ABC):
    """Abstract class for handling different file formats."""

    @abstractmethod
    def load(self, path: Path) -> Dict[str, Any]:
        """Load the configuration file into a dict."""

    @abstractmethod
    def dump(
        self,
        path: Path,
        config: Dict[str, Any],
    ) -> None:
        """Dump in the configuration file."""


C = TypeVar("C")


class Schema(ABC, Generic[C]):
    """Abstract class for handling different config class types."""

    @abstractmethod
    def handles(self, cls: type) -> bool:
        """Return if a type is handled by this schema."""

    @abstractmethod
    def from_dict(self, cls: Type[C], cfg: Dict[str, Any]) -> C:
        """Load the schema based on a dict configuration."""

    @abstractmethod
    def to_dict(self, config: Any, encoder: Encoder) -> Dict[str, Any]:
        """Dump the config to dict."""


class BaseConfig:
    """Abstract base class for handling config files."""

    # Environment variable name holding the config file path to load.
    # Override to disable.
    ENV_PATH: ClassVar[Optional[str]] = "CONFIG"
    # Hardcoded config file path to load.
    # Fallback when no path is found in the environment variable.
    PATH: ClassVar[Optional[str]] = None
    # Selected schema class instance.
    # Override to disable auto selection or control dump options.
    SCHEMA: ClassVar[Optional[Schema]] = None

    # Paths of all config files handled.
    __PATHS: ClassVar[Optional[Tuple[Path, ...]]] = None
    # All formats supported, by extension.
    __FORMATS: ClassVar[Dict[str, Format]] = {}
    # All schema classes supported.
    __SCHEMAS: ClassVar[List[Schema]] = []

    @classmethod
    def register_format(cls, fmt: Format, *extensions: str) -> None:
        """Add a format class to the list of supported ones."""
        for ext in extensions:
            cls.__FORMATS[ext] = fmt

    @classmethod
    def register_schema(cls, schema: Schema[C]) -> None:
        """Add a schema class to the list of supported ones."""
        cls.__SCHEMAS.append(schema)

    @classmethod
    def _paths(cls) -> Tuple[Path, ...]:
        """Cached method to get all handled file paths."""
        if cls.__PATHS:
            return cls.__PATHS
        found_path: Optional[str] = None
        if cls.ENV_PATH:
            found_path = os.environ.get(cls.ENV_PATH)
        if not found_path:
            found_path = cls.PATH
        if not found_path:
            raise ZenConfigError(
                f"could not find the config path for config {cls.__qualname__}, tried env variable {cls.ENV_PATH}"
            )
        cls.__PATHS = tuple(
            sorted(_handle_globbing(Path(found_path).expanduser().absolute()))
        )
        return cls.__PATHS

    @classmethod
    def _format(cls, path: Optional[Path] = None) -> Format:
        """Get the format instance for a path."""
        if path:
            suffix = path.suffix
        else:
            paths = cls._paths()
            if len(paths) != 1:
                raise ZenConfigError(
                    "multiple configuration files, use the path parameter"
                )
            suffix = paths[0].suffix
        if suffix not in cls.__FORMATS:
            raise ZenConfigError(
                f"unsupported extension {suffix} for config {cls.__qualname__}, maybe you are missing an extra"
            )
        return cls.__FORMATS[suffix]

    @classmethod
    def _schema(cls) -> Schema:
        """Get the schema instance for this config class."""
        if cls.SCHEMA:
            return cls.SCHEMA
        for schema in cls.__SCHEMAS:
            if not schema.handles(cls):
                continue
            cls.SCHEMA = schema
            return cls.SCHEMA
        raise ZenConfigError(
            f"could not infer config schema for config {cls.__qualname__}, maybe you are missing an extra"
        )


def _handle_globbing(original_path: Path) -> Iterator[Path]:
    """Convert a glob path to all matched paths."""
    directory = Path(original_path)
    glob = False
    while "*" in directory.name or "?" in directory.name or "[" in directory.name:
        directory = directory.parent
        glob = True
    if not glob:
        yield original_path
    else:
        pattern = str(original_path.relative_to(directory))
        for path in directory.rglob(pattern):
            if path.is_file():
                yield path



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

import tomlkit

from zenconfig.base import BaseConfig, Format


@dataclass
class TOMLFormat(Format):
    sort_keys: bool = False

    def load(self, path: Path) -> Dict[str, Any]:
        return tomlkit.loads(path.read_text())

    def dump(
        self,
        path: Path,
        config: Dict[str, Any],
    ) -> None:
        path.write_text(tomlkit.dumps(config, sort_keys=self.sort_keys))


BaseConfig.register_format(TOMLFormat(), ".toml")



from typing import Any, Dict, Type, TypeVar

from typing_extensions import TypeGuard

from zenconfig.base import BaseConfig, Schema
from zenconfig.encoder import Encoder

C = TypeVar("C", bound=dict)


class DictSchema(Schema[C]):
    def handles(self, cls: type) -> TypeGuard[Type[dict]]:
        return issubclass(cls, dict)

    def from_dict(self, cls: Type[C], cfg: Dict[str, Any]) -> C:
        return cls(cfg)

    def to_dict(self, config: C, encoder: Encoder) -> Dict[str, Any]:
        return encoder(config)


BaseConfig.register_schema(DictSchema())



import logging

import zenconfig.formats
import zenconfig.schemas
from zenconfig.base import Format, Schema
from zenconfig.read import MergeStrategy, ReadOnlyConfig
from zenconfig.write import Config

logging.getLogger(__name__).addHandler(logging.NullHandler())
